# -----------------------------------------------------------------------------
# Copyright © 2009- The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------

"""Provides QtOpenGL classes and functions."""

import contextlib

from . import PYQT5, PYQT6, PYSIDE2, PYSIDE6

if PYQT5:
    from PyQt5.QtGui import (
        QOpenGLBuffer,
        QOpenGLContext,
        QOpenGLContextGroup,
        QOpenGLDebugLogger,
        QOpenGLDebugMessage,
        QOpenGLFramebufferObject,
        QOpenGLFramebufferObjectFormat,
        QOpenGLPixelTransferOptions,
        QOpenGLShader,
        QOpenGLShaderProgram,
        QOpenGLTexture,
        QOpenGLTextureBlitter,
        QOpenGLVersionProfile,
        QOpenGLVertexArrayObject,
        QOpenGLWindow,
    )
    from PyQt5.QtOpenGL import *

    # These are not present on some architectures such as armhf
    with contextlib.suppress(ImportError):
        from PyQt5.QtGui import QOpenGLTimeMonitor, QOpenGLTimerQuery

elif PYQT6:
    from PyQt6.QtGui import QOpenGLContext, QOpenGLContextGroup
    from PyQt6.QtOpenGL import *
elif PYSIDE6:
    from PySide6.QtGui import QOpenGLContext, QOpenGLContextGroup
    from PySide6.QtOpenGL import *
elif PYSIDE2:
    from PySide2.QtGui import (
        QOpenGLBuffer,
        QOpenGLContext,
        QOpenGLContextGroup,
        QOpenGLDebugLogger,
        QOpenGLDebugMessage,
        QOpenGLFramebufferObject,
        QOpenGLFramebufferObjectFormat,
        QOpenGLPixelTransferOptions,
        QOpenGLShader,
        QOpenGLShaderProgram,
        QOpenGLTexture,
        QOpenGLTextureBlitter,
        QOpenGLVersionProfile,
        QOpenGLVertexArrayObject,
        QOpenGLWindow,
    )
    from PySide2.QtOpenGL import *

    # These are not present on some architectures such as armhf
    with contextlib.suppress(ImportError):
        from PySide2.QtGui import QOpenGLTimeMonitor, QOpenGLTimerQuery



import pytest

from qtpy import PYQT6, PYSIDE6


@pytest.mark.skipif(PYQT6, reason="Not complete in PyQt6")
@pytest.mark.skipif(PYSIDE6, reason="Not complete in PySide6")
def test_qt3dcore():
    """Test the qtpy.Qt3DCore namespace"""
    Qt3DCore = pytest.importorskip("qtpy.Qt3DCore")

    assert Qt3DCore.QPropertyValueAddedChange is not None
    assert Qt3DCore.QSkeletonLoader is not None
    assert Qt3DCore.QPropertyNodeRemovedChange is not None
    assert Qt3DCore.QPropertyUpdatedChange is not None
    assert Qt3DCore.QAspectEngine is not None
    assert Qt3DCore.QPropertyValueAddedChangeBase is not None
    assert Qt3DCore.QStaticPropertyValueRemovedChangeBase is not None
    assert Qt3DCore.QPropertyNodeAddedChange is not None
    assert Qt3DCore.QDynamicPropertyUpdatedChange is not None
    assert Qt3DCore.QStaticPropertyUpdatedChangeBase is not None
    assert Qt3DCore.ChangeFlags is not None
    assert Qt3DCore.QAbstractAspect is not None
    assert Qt3DCore.QBackendNode is not None
    assert Qt3DCore.QTransform is not None
    assert Qt3DCore.QPropertyUpdatedChangeBase is not None
    assert Qt3DCore.QNodeId is not None
    assert Qt3DCore.QJoint is not None
    assert Qt3DCore.QSceneChange is not None
    assert Qt3DCore.QNodeIdTypePair is not None
    assert Qt3DCore.QAbstractSkeleton is not None
    assert Qt3DCore.QComponentRemovedChange is not None
    assert Qt3DCore.QComponent is not None
    assert Qt3DCore.QEntity is not None
    assert Qt3DCore.QNodeCommand is not None
    assert Qt3DCore.QNode is not None
    assert Qt3DCore.QPropertyValueRemovedChange is not None
    assert Qt3DCore.QPropertyValueRemovedChangeBase is not None
    assert Qt3DCore.QComponentAddedChange is not None
    assert Qt3DCore.QNodeCreatedChangeBase is not None
    assert Qt3DCore.QNodeDestroyedChange is not None
    assert Qt3DCore.QArmature is not None
    assert Qt3DCore.QStaticPropertyValueAddedChangeBase is not None
    assert Qt3DCore.ChangeFlag is not None
    assert Qt3DCore.QSkeleton is not None



import pytest

from qtpy import PYQT6, PYSIDE6


@pytest.mark.skipif(PYQT6, reason="Not complete in PyQt6")
@pytest.mark.skipif(PYSIDE6, reason="Not complete in PySide6")
def test_qt3drender():
    """Test the qtpy.Qt3DRender namespace"""
    Qt3DRender = pytest.importorskip("qtpy.Qt3DRender")

    assert Qt3DRender.QPointSize is not None
    assert Qt3DRender.QFrustumCulling is not None
    assert Qt3DRender.QPickPointEvent is not None
    assert Qt3DRender.QRenderPassFilter is not None
    assert Qt3DRender.QMesh is not None
    assert Qt3DRender.QRayCaster is not None
    assert Qt3DRender.QStencilMask is not None
    assert Qt3DRender.QPickLineEvent is not None
    assert Qt3DRender.QPickTriangleEvent is not None
    assert Qt3DRender.QRenderState is not None
    assert Qt3DRender.QTextureWrapMode is not None
    assert Qt3DRender.QRenderPass is not None
    assert Qt3DRender.QGeometryRenderer is not None
    assert Qt3DRender.QAttribute is not None
    assert Qt3DRender.QStencilOperation is not None
    assert Qt3DRender.QScissorTest is not None
    assert Qt3DRender.QTextureCubeMapArray is not None
    assert Qt3DRender.QRenderTarget is not None
    assert Qt3DRender.QStencilTest is not None
    assert Qt3DRender.QTextureData is not None
    assert Qt3DRender.QBuffer is not None
    assert Qt3DRender.QLineWidth is not None
    assert Qt3DRender.QLayer is not None
    assert Qt3DRender.QTextureRectangle is not None
    assert Qt3DRender.QRenderTargetSelector is not None
    assert Qt3DRender.QPickingSettings is not None
    assert Qt3DRender.QCullFace is not None
    assert Qt3DRender.QAbstractFunctor is not None
    assert Qt3DRender.PropertyReaderInterface is not None
    assert Qt3DRender.QMaterial is not None
    assert Qt3DRender.QAlphaCoverage is not None
    assert Qt3DRender.QClearBuffers is not None
    assert Qt3DRender.QAlphaTest is not None
    assert Qt3DRender.QStencilOperationArguments is not None
    assert Qt3DRender.QTexture2DMultisample is not None
    assert Qt3DRender.QLevelOfDetailSwitch is not None
    assert Qt3DRender.QRenderStateSet is not None
    assert Qt3DRender.QViewport is not None
    assert Qt3DRender.QObjectPicker is not None
    assert Qt3DRender.QPolygonOffset is not None
    assert Qt3DRender.QRenderSettings is not None
    assert Qt3DRender.QFrontFace is not None
    assert Qt3DRender.QTexture3D is not None
    assert Qt3DRender.QTextureBuffer is not None
    assert Qt3DRender.QTechniqueFilter is not None
    assert Qt3DRender.QLayerFilter is not None
    assert Qt3DRender.QFilterKey is not None
    assert Qt3DRender.QRenderSurfaceSelector is not None
    assert Qt3DRender.QEnvironmentLight is not None
    assert Qt3DRender.QMemoryBarrier is not None
    assert Qt3DRender.QNoDepthMask is not None
    assert Qt3DRender.QBlitFramebuffer is not None
    assert Qt3DRender.QGraphicsApiFilter is not None
    assert Qt3DRender.QAbstractTexture is not None
    assert Qt3DRender.QRenderCaptureReply is not None
    assert Qt3DRender.QAbstractLight is not None
    assert Qt3DRender.QAbstractRayCaster is not None
    assert Qt3DRender.QDirectionalLight is not None
    assert Qt3DRender.QDispatchCompute is not None
    assert Qt3DRender.QBufferDataGenerator is not None
    assert Qt3DRender.QPointLight is not None
    assert Qt3DRender.QStencilTestArguments is not None
    assert Qt3DRender.QTexture1D is not None
    assert Qt3DRender.QCameraSelector is not None
    assert Qt3DRender.QProximityFilter is not None
    assert Qt3DRender.QTexture1DArray is not None
    assert Qt3DRender.QBlendEquation is not None
    assert Qt3DRender.QTextureImageDataGenerator is not None
    assert Qt3DRender.QSpotLight is not None
    assert Qt3DRender.QEffect is not None
    assert Qt3DRender.QSeamlessCubemap is not None
    assert Qt3DRender.QTexture2DMultisampleArray is not None
    assert Qt3DRender.QComputeCommand is not None
    assert Qt3DRender.QFrameGraphNode is not None
    assert Qt3DRender.QSortPolicy is not None
    assert Qt3DRender.QTextureImageData is not None
    assert Qt3DRender.QCamera is not None
    assert Qt3DRender.QGeometry is not None
    assert Qt3DRender.QScreenRayCaster is not None
    assert Qt3DRender.QClipPlane is not None
    assert Qt3DRender.QMultiSampleAntiAliasing is not None
    assert Qt3DRender.QRayCasterHit is not None
    assert Qt3DRender.QAbstractTextureImage is not None
    assert Qt3DRender.QNoDraw is not None
    assert Qt3DRender.QPickEvent is not None
    assert Qt3DRender.QRenderCapture is not None
    assert Qt3DRender.QDepthTest is not None
    assert Qt3DRender.QParameter is not None
    assert Qt3DRender.QLevelOfDetail is not None
    assert Qt3DRender.QGeometryFactory is not None
    assert Qt3DRender.QTexture2D is not None
    assert Qt3DRender.QRenderAspect is not None
    assert Qt3DRender.QPaintedTextureImage is not None
    assert Qt3DRender.QDithering is not None
    assert Qt3DRender.QTextureGenerator is not None
    assert Qt3DRender.QBlendEquationArguments is not None
    assert Qt3DRender.QLevelOfDetailBoundingSphere is not None
    assert Qt3DRender.QColorMask is not None
    assert Qt3DRender.QSceneLoader is not None
    assert Qt3DRender.QTextureLoader is not None
    assert Qt3DRender.QShaderProgram is not None
    assert Qt3DRender.QTextureCubeMap is not None
    assert Qt3DRender.QTexture2DArray is not None
    assert Qt3DRender.QTextureImage is not None
    assert Qt3DRender.QCameraLens is not None
    assert Qt3DRender.QRenderTargetOutput is not None
    assert Qt3DRender.QShaderProgramBuilder is not None
    assert Qt3DRender.QTechnique is not None
    assert Qt3DRender.QShaderData is not None



"""Test QtCore."""

import enum
import sys
from datetime import date, datetime, time

import pytest
from packaging.version import parse

from qtpy import (
    PYQT5,
    PYQT6,
    PYQT_VERSION,
    PYSIDE2,
    PYSIDE_VERSION,
    QtCore,
)

_now = datetime.now()
# Make integer milliseconds; `floor` here, don't `round`!
NOW = _now.replace(microsecond=(_now.microsecond // 1000 * 1000))


def test_qtmsghandler():
    """Test qtpy.QtMsgHandler"""
    assert QtCore.qInstallMessageHandler is not None


@pytest.mark.parametrize("method", ["toPython", "toPyDateTime"])
def test_QDateTime_toPython_and_toPyDateTime(method):
    """Test `QDateTime.toPython` and `QDateTime.toPyDateTime`"""
    q_datetime = QtCore.QDateTime(NOW)
    py_datetime = getattr(q_datetime, method)()
    assert isinstance(py_datetime, datetime)
    assert py_datetime == NOW


@pytest.mark.parametrize("method", ["toPython", "toPyDate"])
def test_QDate_toPython_and_toPyDate(method):
    """Test `QDate.toPython` and `QDate.toPyDate`"""
    q_date = QtCore.QDateTime(NOW).date()
    py_date = getattr(q_date, method)()
    assert isinstance(py_date, date)
    assert py_date == NOW.date()


@pytest.mark.parametrize("method", ["toPython", "toPyTime"])
def test_QTime_toPython_and_toPyTime(method):
    """Test `QTime.toPython` and `QTime.toPyTime`"""
    q_time = QtCore.QDateTime(NOW).time()
    py_time = getattr(q_time, method)()
    assert isinstance(py_time, time)
    assert py_time == NOW.time()


def test_qeventloop_exec(qtbot):
    """Test `QEventLoop.exec_` and `QEventLoop.exec`"""
    assert QtCore.QEventLoop.exec_ is not None
    assert QtCore.QEventLoop.exec is not None
    event_loop = QtCore.QEventLoop(None)
    QtCore.QTimer.singleShot(100, event_loop.quit)
    event_loop.exec_()
    QtCore.QTimer.singleShot(100, event_loop.quit)
    event_loop.exec()


def test_qthread_exec():
    """Test `QThread.exec_` and `QThread.exec_`"""
    assert QtCore.QThread.exec_ is not None
    assert QtCore.QThread.exec is not None


@pytest.mark.skipif(
    PYSIDE2 and parse(PYSIDE_VERSION) < parse("5.15"),
    reason="QEnum macro doesn't seem to be present on PySide2 <5.15",
)
def test_qenum():
    """Test QEnum macro"""

    class EnumTest(QtCore.QObject):
        class Position(enum.IntEnum):
            West = 0
            North = 1
            South = 2
            East = 3

        QtCore.QEnum(Position)

    obj = EnumTest()
    assert obj.metaObject().enumerator(0).name() == "Position"


def test_QLibraryInfo_location_and_path():
    """Test `QLibraryInfo.location` and `QLibraryInfo.path`"""
    assert QtCore.QLibraryInfo.location is not None
    assert (
        QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.PrefixPath)
        is not None
    )
    assert QtCore.QLibraryInfo.path is not None
    assert QtCore.QLibraryInfo.path(QtCore.QLibraryInfo.PrefixPath) is not None


def test_QLibraryInfo_LibraryLocation_and_LibraryPath():
    """Test `QLibraryInfo.LibraryLocation` and `QLibraryInfo.LibraryPath`"""
    assert QtCore.QLibraryInfo.LibraryLocation is not None
    assert QtCore.QLibraryInfo.LibraryPath is not None


def test_QCoreApplication_exec_(qapp):
    """Test `QtCore.QCoreApplication.exec_`"""
    assert QtCore.QCoreApplication.exec_ is not None
    app = QtCore.QCoreApplication.instance() or QtCore.QCoreApplication(
        [sys.executable, __file__],
    )
    assert app is not None
    QtCore.QTimer.singleShot(100, QtCore.QCoreApplication.instance().quit)
    QtCore.QCoreApplication.exec_()
    app = QtCore.QCoreApplication.instance() or QtCore.QCoreApplication(
        [sys.executable, __file__],
    )
    assert app is not None
    QtCore.QTimer.singleShot(100, QtCore.QCoreApplication.instance().quit)
    app.exec_()


def test_QCoreApplication_exec(qapp):
    """Test `QtCore.QCoreApplication.exec`"""
    assert QtCore.QCoreApplication.exec is not None
    app = QtCore.QCoreApplication.instance() or QtCore.QCoreApplication(
        [sys.executable, __file__],
    )
    assert app is not None
    QtCore.QTimer.singleShot(100, QtCore.QCoreApplication.instance().quit)
    QtCore.QCoreApplication.exec()
    app = QtCore.QCoreApplication.instance() or QtCore.QCoreApplication(
        [sys.executable, __file__],
    )
    assert app is not None
    QtCore.QTimer.singleShot(100, QtCore.QCoreApplication.instance().quit)
    app.exec()


@pytest.mark.skipif(
    PYQT5 or PYQT6,
    reason="Doesn't seem to be present on PyQt5 and PyQt6",
)
def test_qtextstreammanipulator_exec():
    """Test `QTextStreamManipulator.exec_` and `QTextStreamManipulator.exec`"""
    assert QtCore.QTextStreamManipulator.exec_ is not None
    assert QtCore.QTextStreamManipulator.exec is not None


@pytest.mark.skipif(
    PYSIDE2 or PYQT6,
    reason="Doesn't seem to be present on PySide2 and PyQt6",
)
def test_QtCore_SignalInstance():
    class ClassWithSignal(QtCore.QObject):
        signal = QtCore.Signal()

    instance = ClassWithSignal()

    assert isinstance(instance.signal, QtCore.SignalInstance)


@pytest.mark.skipif(
    PYQT5 and PYQT_VERSION.startswith("5.9"),
    reason="A specific setup with at least sip 4.9.9 is needed for PyQt5 5.9.*"
    "to work with scoped enum access",
)
def test_enum_access():
    """Test scoped and unscoped enum access for qtpy.QtCore.*."""
    assert (
        QtCore.QAbstractAnimation.Stopped
        == QtCore.QAbstractAnimation.State.Stopped
    )
    assert QtCore.QEvent.ActionAdded == QtCore.QEvent.Type.ActionAdded
    assert QtCore.Qt.AlignLeft == QtCore.Qt.AlignmentFlag.AlignLeft
    assert QtCore.Qt.Key_Return == QtCore.Qt.Key.Key_Return
    assert QtCore.Qt.transparent == QtCore.Qt.GlobalColor.transparent
    assert QtCore.Qt.Widget == QtCore.Qt.WindowType.Widget
    assert QtCore.Qt.BackButton == QtCore.Qt.MouseButton.BackButton
    assert QtCore.Qt.XButton1 == QtCore.Qt.MouseButton.XButton1
    assert (
        QtCore.Qt.BackgroundColorRole
        == QtCore.Qt.ItemDataRole.BackgroundColorRole
    )
    assert QtCore.Qt.TextColorRole == QtCore.Qt.ItemDataRole.TextColorRole
    assert QtCore.Qt.MidButton == QtCore.Qt.MouseButton.MiddleButton


@pytest.mark.skipif(
    PYSIDE2 and PYSIDE_VERSION.startswith("5.12.0"),
    reason="Utility functions unavailable for PySide2 5.12.0",
)
def test_qtgui_namespace_mightBeRichText():
    """
    Test included elements (mightBeRichText) from module QtGui.

    See: https://doc.qt.io/qt-5/qt-sub-qtgui.html
    """
    assert QtCore.Qt.mightBeRichText is not None


def test_itemflags_typedef():
    """
    Test existence of `QFlags<ItemFlag>` typedef `ItemFlags` that was removed from PyQt6
    """
    assert QtCore.Qt.ItemFlags is not None
    assert QtCore.Qt.ItemFlags() == QtCore.Qt.ItemFlag(0)



import pytest

from qtpy import PYSIDE2


@pytest.mark.skipif(PYSIDE2, reason="Not available in CI")
def test_qtserialport():
    """Test the qtpy.QtSerialPort namespace"""
    QtSerialPort = pytest.importorskip("qtpy.QtSerialPort")

    assert QtSerialPort.QSerialPort is not None
    assert QtSerialPort.QSerialPortInfo is not None



from qtpy import PYQT5, PYSIDE2


def test_qtquick():
    """Test the qtpy.QtQuick namespace"""
    from qtpy import QtQuick

    if PYQT5:
        assert QtQuick.QQuickCloseEvent is not None
        assert QtQuick.QSGFlatColorMaterial is not None
        assert QtQuick.QSGImageNode is not None
        assert QtQuick.QSGMaterial is not None
        assert QtQuick.QSGMaterialShader is not None
        assert QtQuick.QSGOpaqueTextureMaterial is not None
        assert QtQuick.QSGRectangleNode is not None
        assert QtQuick.QSGRenderNode is not None
        assert QtQuick.QSGRendererInterface is not None
        assert QtQuick.QSGTextureMaterial is not None
        assert QtQuick.QSGVertexColorMaterial is not None

    assert QtQuick.QQuickAsyncImageProvider is not None
    assert QtQuick.QQuickFramebufferObject is not None
    assert QtQuick.QQuickImageProvider is not None
    assert QtQuick.QQuickImageResponse is not None
    assert QtQuick.QQuickItem is not None
    assert QtQuick.QQuickItemGrabResult is not None
    assert QtQuick.QQuickPaintedItem is not None
    assert QtQuick.QQuickRenderControl is not None
    assert QtQuick.QQuickTextDocument is not None
    assert QtQuick.QQuickTextureFactory is not None
    assert QtQuick.QQuickView is not None
    assert QtQuick.QQuickWindow is not None
    if PYQT5 or PYSIDE2:
        assert QtQuick.QSGAbstractRenderer is not None
        assert QtQuick.QSGEngine is not None
    assert QtQuick.QSGBasicGeometryNode is not None
    assert QtQuick.QSGClipNode is not None
    assert QtQuick.QSGDynamicTexture is not None
    assert QtQuick.QSGGeometry is not None
    assert QtQuick.QSGGeometryNode is not None
    assert QtQuick.QSGMaterialType is not None
    assert QtQuick.QSGNode is not None
    assert QtQuick.QSGOpacityNode is not None
    assert QtQuick.QSGSimpleRectNode is not None
    assert QtQuick.QSGSimpleTextureNode is not None
    assert QtQuick.QSGTexture is not None
    assert QtQuick.QSGTextureProvider is not None
    assert QtQuick.QSGTransformNode is not None



import pytest

from qtpy import PYSIDE2, PYSIDE6


@pytest.mark.skipif(
    not (PYSIDE2 or PYSIDE6),
    reason="Only available by default in PySide",
)
def test_qtcharts():
    """Test the qtpy.QtCharts namespace"""
    QtCharts = pytest.importorskip("qtpy.QtCharts")

    assert QtCharts.QChart is not None
    assert QtCharts.QtCharts.QChart is not None



import pytest


def test_qtwebenginecore():
    """Test the qtpy.QtWebEngineCore namespace"""
    QtWebEngineCore = pytest.importorskip("qtpy.QtWebEngineCore")

    assert QtWebEngineCore.QWebEngineHttpRequest is not None



import pytest


def test_qt3danimation():
    """Test the qtpy.Qt3DAnimation namespace"""
    Qt3DAnimation = pytest.importorskip("qtpy.Qt3DAnimation")

    assert Qt3DAnimation.QAnimationController is not None
    assert Qt3DAnimation.QAdditiveClipBlend is not None
    assert Qt3DAnimation.QAbstractClipBlendNode is not None
    assert Qt3DAnimation.QAbstractAnimation is not None
    assert Qt3DAnimation.QKeyframeAnimation is not None
    assert Qt3DAnimation.QAbstractAnimationClip is not None
    assert Qt3DAnimation.QAbstractClipAnimator is not None
    assert Qt3DAnimation.QClipAnimator is not None
    assert Qt3DAnimation.QAnimationGroup is not None
    assert Qt3DAnimation.QLerpClipBlend is not None
    assert Qt3DAnimation.QMorphingAnimation is not None
    assert Qt3DAnimation.QAnimationAspect is not None
    assert Qt3DAnimation.QVertexBlendAnimation is not None
    assert Qt3DAnimation.QBlendedClipAnimator is not None
    assert Qt3DAnimation.QMorphTarget is not None



# -----------------------------------------------------------------------------
# Copyright © 2009- The Spyder Development Team
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)
# -----------------------------------------------------------------------------

"""Provides Qsci classes and functions."""

from . import (
    PYQT5,
    PYQT6,
    PYSIDE2,
    PYSIDE6,
    QtBindingMissingModuleError,
    QtModuleNotInstalledError,
)

if PYQT5:
    try:
        from PyQt5.Qsci import *
    except ModuleNotFoundError as error:
        raise QtModuleNotInstalledError(
            name="Qsci",
            missing_package="QScintilla",
        ) from error
elif PYQT6:
    try:
        from PyQt6.Qsci import *
    except ModuleNotFoundError as error:
        raise QtModuleNotInstalledError(
            name="Qsci",
            missing_package="PyQt6-QScintilla",
        ) from error
elif PYSIDE2 or PYSIDE6:
    raise QtBindingMissingModuleError(name="Qsci")



"""Test QtGui."""

import sys

import pytest

from qtpy import (
    PYQT5,
    PYQT_VERSION,
    PYSIDE2,
    PYSIDE6,
    QtCore,
    QtGui,
    QtWidgets,
)
from qtpy.tests.utils import not_using_conda


def test_qfontmetrics_width(qtbot):
    """Test QFontMetrics and QFontMetricsF width"""
    assert QtGui.QFontMetrics.width is not None
    assert QtGui.QFontMetricsF.width is not None
    font = QtGui.QFont("times", 24)
    font_metrics = QtGui.QFontMetrics(font)
    font_metricsF = QtGui.QFontMetricsF(font)
    width = font_metrics.width("Test")
    widthF = font_metricsF.width("Test")
    assert width in range(40, 62)
    assert 39 <= widthF <= 63


def test_qdrag_functions(qtbot):
    """Test functions mapping for QtGui.QDrag."""
    assert QtGui.QDrag.exec_ is not None
    drag = QtGui.QDrag(None)
    drag.exec_()


def test_QGuiApplication_exec_():
    """Test `QtGui.QGuiApplication.exec_`"""
    assert QtGui.QGuiApplication.exec_ is not None
    app = QtGui.QGuiApplication.instance() or QtGui.QGuiApplication(
        [sys.executable, __file__],
    )
    assert app is not None
    QtCore.QTimer.singleShot(100, QtGui.QGuiApplication.instance().quit)
    QtGui.QGuiApplication.exec_()
    app = QtGui.QGuiApplication.instance() or QtGui.QGuiApplication(
        [sys.executable, __file__],
    )
    assert app is not None
    QtCore.QTimer.singleShot(100, QtGui.QGuiApplication.instance().quit)
    app.exec_()


def test_what_moved_to_qtgui_in_qt6():
    """Test what has been moved to QtGui in Qt6"""
    assert QtGui.QAction is not None
    assert QtGui.QActionGroup is not None
    assert QtGui.QFileSystemModel is not None
    assert QtGui.QShortcut is not None
    assert QtGui.QUndoCommand is not None


def test_qtextdocument_functions(pdf_writer):
    """Test functions mapping for QtGui.QTextDocument."""
    assert QtGui.QTextDocument.print_ is not None
    text_document = QtGui.QTextDocument("Test")
    print_device, output_path = pdf_writer
    text_document.print_(print_device)
    assert output_path.exists()


@pytest.mark.skipif(
    PYQT5 and PYQT_VERSION.startswith("5.9"),
    reason="A specific setup with at least sip 4.9.9 is needed for PyQt5 5.9.*"
    "to work with scoped enum access",
)
def test_enum_access():
    """Test scoped and unscoped enum access for qtpy.QtWidgets.*."""
    assert QtGui.QColor.Rgb == QtGui.QColor.Spec.Rgb
    assert QtGui.QFont.AllUppercase == QtGui.QFont.Capitalization.AllUppercase
    assert QtGui.QIcon.Normal == QtGui.QIcon.Mode.Normal
    assert QtGui.QImage.Format_Invalid == QtGui.QImage.Format.Format_Invalid


@pytest.mark.skipif(
    sys.platform == "darwin" and sys.version_info[:2] == (3, 7),
    reason="Stalls on macOS CI with Python 3.7",
)
def test_QSomethingEvent_pos_functions(qtbot):
    """
    Test `QMouseEvent.pos` and related functions removed in Qt 6,
    and `QMouseEvent.position`, etc., missing from Qt 5.
    """

    class Window(QtWidgets.QMainWindow):
        def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent) -> None:
            assert event.globalPos() - event.pos() == self.mapToParent(
                QtCore.QPoint(0, 0),
            )
            assert event.pos().x() == event.x()
            assert event.pos().y() == event.y()
            assert event.globalPos().x() == event.globalX()
            assert event.globalPos().y() == event.globalY()
            assert event.position().x() == event.pos().x()
            assert event.position().y() == event.pos().y()
            assert event.globalPosition().x() == event.globalPos().x()
            assert event.globalPosition().y() == event.globalPos().y()

            event.accept()

    window = Window()
    window.setMinimumSize(320, 240)  # ensure the window is of sufficient size
    window.show()

    with qtbot.waitExposed(window):
        qtbot.mouseMove(window, QtCore.QPoint(42, 6 * 9))
        qtbot.mouseDClick(window, QtCore.Qt.LeftButton)

    # the rest of the functions are not actually tested
    # QSinglePointEvent (Qt6) child classes checks
    for _class in ("QNativeGestureEvent", "QEnterEvent", "QTabletEvent"):
        for _function in (
            "pos",
            "x",
            "y",
            "globalPos",
            "globalX",
            "globalY",
            "position",
            "globalPosition",
        ):
            assert hasattr(getattr(QtGui, _class), _function)

    # QHoverEvent checks
    for _function in ("pos", "x", "y", "position"):
        assert hasattr(QtGui.QHoverEvent, _function)

    # QDropEvent and child classes checks
    for _class in ("QDropEvent", "QDragMoveEvent", "QDragEnterEvent"):
        for _function in ("pos", "posF", "position"):
            assert hasattr(getattr(QtGui, _class), _function)


@pytest.mark.skipif(
    not (PYSIDE2 or PYSIDE6),
    reason="PySide{2,6} specific test",
)
def test_qtextcursor_moveposition():
    """Test monkeypatched QTextCursor.movePosition"""
    doc = QtGui.QTextDocument("foo bar baz")
    cursor = QtGui.QTextCursor(doc)

    assert not cursor.movePosition(QtGui.QTextCursor.Start)
    assert cursor.movePosition(
        QtGui.QTextCursor.EndOfWord,
        mode=QtGui.QTextCursor.KeepAnchor,
    )
    assert cursor.selectedText() == "foo"

    assert cursor.movePosition(QtGui.QTextCursor.Start)
    assert cursor.movePosition(
        QtGui.QTextCursor.WordRight,
        n=2,
        mode=QtGui.QTextCursor.KeepAnchor,
    )
    assert cursor.selectedText() == "foo bar "

    assert cursor.movePosition(QtGui.QTextCursor.Start)
    assert cursor.position() == cursor.anchor()
    assert cursor.movePosition(
        QtGui.QTextCursor.NextWord,
        QtGui.QTextCursor.KeepAnchor,
        3,
    )
    assert cursor.selectedText() == "foo bar baz"


def test_opengl_imports():
    """
    Test for presence of QOpenGL* classes.

    These classes were members of QtGui in Qt5, but moved to QtOpenGL in Qt6.
    QtPy makes them available in QtGui to maintain compatibility.
    """

    assert QtGui.QOpenGLBuffer is not None
    assert QtGui.QOpenGLContext is not None
    assert QtGui.QOpenGLContextGroup is not None
    assert QtGui.QOpenGLDebugLogger is not None
    assert QtGui.QOpenGLDebugMessage is not None
    assert QtGui.QOpenGLFramebufferObject is not None
    assert QtGui.QOpenGLFramebufferObjectFormat is not None
    assert QtGui.QOpenGLPixelTransferOptions is not None
    assert QtGui.QOpenGLShader is not None
    assert QtGui.QOpenGLShaderProgram is not None
    assert QtGui.QOpenGLTexture is not None
    assert QtGui.QOpenGLTextureBlitter is not None
    assert QtGui.QOpenGLVersionProfile is not None
    assert QtGui.QOpenGLVertexArrayObject is not None
    assert QtGui.QOpenGLWindow is not None
